# Demonstrate usage of PyMa27 abstract class
# for the solution of symmetric systems of
# linear equations.

import pyma27
from pysparse import spmatrix
import numpy
from string import strip, atof
from nlpy_timing import cputime

def SolveSystem( A, rhs ):
    t = cputime()
    P = pyma27.PyMa27Context( A )
    t_analyze = cputime() - t
    t = cputime()
    P.solve( rhs, get_resid = False )   # Skip computation of residual
    t_solve = cputime() - t

    return (P.x, t_analyze, t_solve)

if __name__ == '__main__':
    import sys
    import os
    matrices = sys.argv[1:]

    fmt_head = '%-15s  %5s  %7s  %8s  %8s  %7s  %7s'
    fmt_data = '%-15s  %5d  %7d  %8.1e  %8.1e  %7.2f  %7.2f'
    header = fmt_head % ('Name','Order','Nnz','Resid','Error','Analyze','Solve')
    header_len = len( header )
    sys.stderr.write( '-' * header_len + '\n' )
    sys.stderr.write( header + '\n' )
    sys.stderr.write( '-' * header_len + '\n' )

    analyze_python = {}
    solve_python = {}
    ProbList = []
    # Process matrices given on the command line
    for matrix in matrices:
        M = spmatrix.ll_mat_from_mtx( matrix )
        (m,n) = M.shape
        if m != n: break
        e = numpy.ones( n, 'd' )
        rhs = numpy.empty( n, 'd' )
        M.matvec( e, rhs )
        (sol, t_an, t_sl) = SolveSystem( M, rhs )
        err = numpy.linalg.norm( sol - e, ord=numpy.Inf )
        res = numpy.empty( n, 'd' )
        M.matvec( sol, res )
        relres = numpy.linalg.norm( res - rhs, ord=numpy.Inf )
        relres /= max( 1.0, numpy.linalg.norm( rhs, ord=numpy.Inf ) )
        probname = os.path.basename( matrix )
        if probname[-4:] == '.mtx': probname = probname[:-4]
	ProbList.append( probname )
	analyze_python[probname] = t_an
	solve_python[probname] = t_sl
        sys.stdout.write( fmt_data % (probname,n,M.nnz,err,relres,t_an,t_sl) )
        sys.stdout.write( '\n' )
    sys.stderr.write( '-' * header_len + '\n' )

    ProbList.sort()
    nprobs = len( ProbList )
        
    # Obtain timings from pure Fortran and C interface to Ma27
    # These timings were obtained using the built-in dtime() and clock()
    # with the same MA27 options
    try:
        ft = open( '../.benchmark/Ma27/ma27_timings', 'r' )
    except:
        sys.stderr.write( 'Cannot open ../.benchmark/Ma27/ma27_timings\n' )
        sys.stderr.write( 'Make sure you have run "make bmark"\n' )
        sys.exit(1)

    f77data = ft.readlines()
    ft.close()
    analyze_f77 = {}
    solve_f77 = {}
    for line in f77data:
        line = line.split()
        pbname = strip( line[0] )
	if pbname[-4:] == '.mtx': pbname = pbname[:-4]
        analyze_f77[pbname] = atof(line[1])
        solve_f77[pbname] = atof(line[2])

    try:
        ct = open( '../.benchmark/Ma27/ma27_c_timings', 'r' )
    except:
        sys.stderr.write( 'Cannot open ../.benchmark/Ma27/ma27_c_timings\n' )
        sys.stderr.write( 'Make sure you have run "make bmark"\n' )
        sys.exit(1)

    cdata = ct.readlines()
    ct.close()
    analyze_c = {}
    solve_c = {}
    for line in cdata:
        line = line.split()
        pbname = strip( line[0] )
	if pbname[-4:] == '.mtx': pbname = pbname[:-4]
        analyze_c[pbname] = atof(line[1])
        solve_c[pbname] = atof(line[2])

    # Plot the timings
    try:
        import matplotlib
        if matplotlib.__version__ < '0.65':
            import matplotlib.matlab as MM
        else:
            import matplotlib.pylab as MM
    except:
        sys.stderr.write( ' If you had Matplotlib installed, you\n' )
        sys.stderr.write( ' would be looking at timing plots right now.\n' )
        sys.exit( 0 )

    darkblue = '#2c11cf'
    lightblue = '#8f84e0'
    steelblue = '#5d82ef'
    x = range( nprobs )
    fig = MM.figure()
    ax1 = fig.add_subplot(211)
    ax1.plot( x, [analyze_f77[k] for k in ProbList],  color = darkblue,  linewidth = 3 )
    ax1.plot( x, [analyze_c[k] for k in ProbList], color = lightblue, linewidth = 3 )
    ax1.plot( x, [analyze_python[k] for k in ProbList], color = steelblue, linewidth = 3 )
    ax1.set_xticks([])
    ax1.set_ylabel( 'Analyze Time (s)' )
    ax1.set_title( 'Comparing Fortran, C and Python interfaces to MA27' )
    ax = fig.add_subplot(212)
    ax.plot( x, [solve_f77[k] for k in ProbList],  color = darkblue,  linewidth = 3 )
    ax.plot( x, [solve_c[k] for k in ProbList], color = lightblue, linewidth = 3 )
    ax.plot( x, [solve_python[k] for k in ProbList], color = steelblue, linewidth = 3 )
    ax.set_xticks( x )
    ax.set_xticklabels( ProbList, rotation=45, horizontalalignment='right' )
    ax.set_ylabel( 'Solve Time (s)' )
    ax.legend( ['Fortran', 'C', 'Python'], 'upper right' )
    MM.show()
