# Demonstrate usage of PyMa27 abstract class
# for the solution of symmetric systems of
# linear equations.

import pyma27
from pysparse import spmatrix
from   sparsetools import Hilbert
import numpy
import norms
from nlpy_timing import cputime

def Ma27SpecSheet():
    # This is the example from the MA27 spec sheet
    # Solution should be [1,2,3,4,5]
    A = spmatrix.ll_mat_sym( 5, 7 )
    A[0,0] = 2
    A[1,0] = 3
    A[2,1] = 4
    A[2,2] = 1
    A[3,2] = 5
    A[4,1] = 6
    A[4,4] = 1

    rhs = numpy.ones( 5, 'd' )
    rhs[0] = 8
    rhs[1] = 45
    rhs[2] = 31
    rhs[3] = 15
    rhs[4] = 17

    return (A, rhs)

def SolveSystem( A, rhs, itref_threshold = 1.0e-6, nitrefmax = 5 ):

    # Obtain pyma27 context object
    t = cputime()
    P = pyma27.PyMa27Context( A )
    t_analyze = cputime() - t

    # Solve system and compute residual
    t = cputime()
    P.solve( rhs )
    t_solve = cputime() - t_analyze

    # Compute residual norm
    nrhsp1 = norms.norm_infty(rhs) + 1
    nr = norms.norm2( P.residual )/nrhsp1

    # If residual not small, perform iterative refinement
    P.refine( rhs, tol = itref_threshold, nitref = nitrefmax )
    nr1 = norms.norm_infty( P.residual )/nrhsp1

    return (P.x, P.residual, nr, nr1, t_analyze, t_solve, P.neig)

if __name__ == '__main__':
    import sys
    import os
    matrices = sys.argv[1:]

    header = '%-13s  %-11s  %-11s  %-11s  %-7s  %-7s  %-5s\n' % ('Name', 'Rel. resid.', 'Residual', 'Resid itref', 'Analyze', 'Solve', 'neig' )
    lhead = len(header)
    sys.stderr.write( '-' * lhead + '\n' )
    sys.stderr.write( header )
    sys.stderr.write( '-' * lhead + '\n' )

    # Solve example from the spec sheet
    (A, rhs) = Ma27SpecSheet()
    (x, r, nr, nr1, t_an, t_sl, neig) = SolveSystem( A, rhs )
    exact = numpy.arange( 5, dtype = 'd' ) + 1
    relres = norms.norm2( x - exact ) / norms.norm2( exact )
    sys.stdout.write( '%-13s  %-11.5e  %-11.5e  %-11.5e  %-7.3f  %-7.3f  %-5d\n' % ('Spec sheet', relres, nr, nr1, t_an, t_sl, neig) )

    # Solve example with Hilbert matrix
    n = 10
    H = Hilbert( n )
    e = numpy.ones( n, 'd' )
    rhs = numpy.zeros( n, 'd' )
    H.matvec( e, rhs )
    (x, r, nr, nr1, t_an, t_sl, neig) = SolveSystem( H, rhs )
    relres = norms.norm2( x - e ) / norms.norm2( e )
    sys.stdout.write( '%-13s  %-11.5e  %-11.5e  %-11.5e  %-7.3f  %-7.3f  %-5d\n' % ('Hilbert', relres, nr, nr1, t_an, t_sl, neig) )

    # Process matrices given on the command line
    for matrix in matrices:
        M = spmatrix.ll_mat_from_mtx( matrix )
        (m,n) = M.shape
        if m != n: break
        e = numpy.ones( n, 'd' )
        rhs = numpy.zeros( n, 'd' )
        M.matvec( e, rhs )
        (x, r, nr, nr1, t_an, t_sl, neig) = SolveSystem( M, rhs )
        relres = norms.norm2( x - e ) / norms.norm2( e )
        probname = os.path.basename( matrix )
        if probname[-4:] == '.mtx': probname = probname[:-4]
        sys.stdout.write( '%-13s  %-11.5e  %-11.5e  %-11.5e  %-7.3f  %-7.3f  %-5d\n' % (probname, relres, nr, nr1, t_an, t_sl, neig) )
    sys.stderr.write( '-' * lhead + '\n' )
        
