# Simple example:
#    Newton's method
#    with Armijo backtracking
# This example uses dense linear algebra

import numpy
from norms import norm_infty, norm2
from linesearch import ArmijoLineSearch
from rosenbrock import RosenbrockModel
import sys

def rosenbrockxy(x,y):
    return RosenbrockModel.obj( (x,y) )
    
ALS = ArmijoLineSearch( tfactor = 0.2 )
x = numpy.array( [-0.5, 1.0], 'd' )
xmin = xmax = x[0]
ymin = ymax = x[1]

# Initialization
f = RosenbrockModel.obj(x)
grad = RosenbrockModel.grad(x)
ngrad = norm_infty(grad)
n2grad = norm2(grad)

t = 0.0
tlist = []
xlist = [x[0]]
ylist = [x[1]]
iter = 0

asciiformat1 = '%-3d  %-12g  %-12g  %-12g  %-12g  '
asciiformat2 = '%-6g  %-12g  %-3d'

texformat1 = '%-3d & %-12g & %-12g & %-12g & %-12g & '
texformat2 = '%-6g & %-12g & %-3d \\\\'

format1 = asciiformat1 # or texformat1 for TeX output
format2 = asciiformat2 # or texformat2 for TeX output

sys.stdout.write( format1 % (iter, f, ngrad, x[0], x[1] ) )

while ngrad > 1.0e-8:

    iter += 1

    # Compute Newton direction
    H = RosenbrockModel.hess(x)
    d = numpy.linalg.solve( H, -grad )
    nd = norm2(d) #norm_infty( d )
    slope = numpy.dot(grad, d)
    cosine = slope/(n2grad * nd)

    # Correct curvature if necessary
    n_corr = 0
    while cosine >= 0:
        n_corr = n_corr + 1
        l = numpy.linalg.eigvals(H)
        lambda_min = min(l.real)
        H = H + (0.1 - lambda_min) * numpy.identity(2, 'd')
        d = numpy.linalg.solve(H, -grad)
        nd = norm2(d) #norm_infty( d )
        cosine = numpy.dot(grad, d)/(n2grad * nd)
        
    sys.stdout.write( format2 % (t, cosine, 0) + '\n' )
        
    # Perform linesearch
    t = ALS.search( RosenbrockModel.obj, x, d, slope, f = f )
    tlist.append(t)

    # Move ahead
    x += t * d
    xlist.append(x[0])
    ylist.append(x[1])
    xmin = min( xmin, x[0] )
    xmax = max( xmax, x[0] )
    ymin = min( ymin, x[1] )
    ymax = max( ymax, x[1] )
    f = RosenbrockModel.obj( x )
    grad = RosenbrockModel.grad( x )
    ngrad = norm_infty( grad )
    n2grad = norm2( grad )
    
    sys.stdout.write( format1 % (iter, f, ngrad, x[0], x[1] ) )

sys.stdout.write( '\n' )

try:
    from matplotlib.pylab import *
except:
    import sys
    sys.stderr.write( 'If you had Matplotlib, you would be looking\n' )
    sys.stderr.write( 'at a contour plot right now...\n' )
    sys.exit(0)

xx = arange( -1.0, 1.1, 0.01 )
yy = arange( -0.1, 1.2, 0.01 )
XX, YY = meshgrid( xx, yy )
ZZ = rosenbrockxy( XX, YY )
plot( xlist, ylist, 'r-', lw = 1.5 )
plot( [xlist[0]], [ylist[0]], 'go', [xlist[-1]], [ylist[-1]], 'go' )
contour( XX, YY, ZZ, linewidths = 1.5, alpha = 0.75, origin = 'lower' )
title( "Newton's method with Armijo linesearch on the Rosenbrock function" )
show()
#savefig( 'newton-rosenbrock.png' )
