import numpy
from pysparse import spmatrix
import pycfs
from nlpy_timing import cputime
import sys

def usage():
    progname = sys.argv[0]
    sys.stderr.write( 'Usage\n' );
    sys.stderr.write( '  %-s problem [problem ...]\n' % progname )
    sys.stderr.write( '  where each problem is the file name of a matrix\n' )
    sys.stderr.write( '  in MatrixMarket sparse format (*.mtx).\n' )

def test_pycfs( ProblemList, plist = [0, 2, 5, 10] ):
    import os

    if len( ProblemList ) == 0:
        usage()
        sys.exit( 1 )
    
    fmt = '%-15s  %-5s  %-6s  %-2s  %-4s  %-9s  %-4s  %-12s  %-6s  %-6s\n'
    header = fmt % ('Name','Size','nnz','p','info','shift','iter','rel resid','fact','solve')
    lhead = len( header )
    sys.stderr.write( '-' * lhead + '\n' )
    sys.stderr.write( header )
    sys.stderr.write( '-' * lhead + '\n' )

    time_list = {}        # Record timings
    iter_list = {}        # Record number of iterations

    for problem in ProblemList:

        A = spmatrix.ll_mat_from_mtx( problem )
        (m, n) = A.shape
        if m != n: break

        prob = os.path.basename( problem )
        if prob[-4:] == '.mtx': prob = prob[:-4]

        # Right-hand side is Ae, as in Icfs.
        e = numpy.ones( n, 'd' );
        b = numpy.ones( n, 'd' );
        A.matvec( e, b )

        sys.stdout.write( '%-15s  %-5d  %-6d  ' % (prob, n, A.nnz) )
        advance = False

        # Call icfs and pcg
        tlist = []
        ilist = []
        for pval in plist:
            t0 = cputime()
            P = pycfs.PycfsContext( A, mem = pval )
            t_fact = cputime() - t0
            P.solve( b )
            t_solve = cputime() - t_fact
            tlist.append( t_fact + t_solve )
            ilist.append( P.iter )

            if advance: sys.stdout.write( ' ' * 32 )
            sys.stdout.write( '%-2d  %-4d  %-9g  %-4d  %-12g  %-6g  %-6g\n' % (pval, P.info, P.shift, P.iter, P.relres, t_fact, t_solve) )
            advance = True
            
        time_list[prob] = tlist
        iter_list[prob] = ilist
        sys.stderr.write( '-' * lhead + '\n' )

    return (time_list,iter_list)


if __name__ == '__main__':
    # Obtain timings for the specified problems
    plist = [0, 2, 5, 10]
    (t_lst, i_lst) = test_pycfs( sys.argv[1:], plist = plist )

    # I obtained time_ref by running the default icf driver on my machine
    # for the same values of p as in plist.
    t_ref = { '1138bus'  : [0.300E-01, 0.200E-01, 0.100E-01, 0.100E-01],
              'bcsstk08' : [0.140E+00, 0.130E+00, 0.300E-01, 0.300E-01],
              'bcsstk09' : [0.200E-01, 0.200E-01, 0.200E-01, 0.900E-01],
              'bcsstk10' : [0.200E-01, 0.200E-01, 0.100E-01, 0.200E-01],
              'bcsstk11' : [0.120E+00, 0.140E+00, 0.130E+00, 0.130E+00],
              'bcsstk18' : [0.720E+00, 0.400E+00, 0.380E+00, 0.340E+00],
              'bcsstk19' : [0.100E-01, 0.100E-01, 0.100E-01, 0.000E+00] }

    K = t_lst.keys()

    if len( K ) > 1:
        try:
            import matplotlib
            if matplotlib.__version__ < '0.65':
                import matplotlib.matlab as MM
            else:
                import matplotlib.pylab as MM
        except:
            print ' If you had Matplotlib installed, you would be looking'
            print ' at timing plots right now...'
            sys.exit( 0 )

        darkblue = '#2c11cf'
        lightblue = '#8f84e0'
        steelblue = '#5d82ef'
        x = range( len( t_lst.keys() ) )
        ax = []
        ax.append( MM.axes( [ .05, .05, .40, .40 ] ) ) # lower left
        ax.append( MM.axes( [ .05, .55, .40, .40 ] ) ) # upper left
        ax.append( MM.axes( [ .55, .05, .40, .40 ] ) ) # lower right
        ax.append( MM.axes( [ .55, .55, .40, .40 ] ) ) # upper right
        for i in range( 4 ):
            ax[i].plot( x, [ t_lst[k][i] for k in K ], color = darkblue,  linewidth = 3 )
            ax[i].plot( x, [ t_ref[k][i] for k in K ], color = lightblue, linewidth = 3 )
            ax[i].set_xticks( [] ) #x )
            ax[i].set_title( 'Limited memory p = %-d' % plist[i], fontsize = 'small' )
            ax[i].legend( ['Python', 'Fortran'], 'upper left' )
        for i in [2,3]:
            ax[i].set_ylabel( 'Time (s)', fontsize = 'small' )
        MM.show()

        # For the number of iterations, use first value of p as reference
        x = range( len( i_lst.keys() ) )
        ax = MM.subplot(111)
        lgnd = []
        for i in range( len( plist ) ):
            lgnd.append( 'p = %-d' % plist[i] )
        ax.plot( x, [ (1.0*i_lst[k][0])/i_lst[k][0] for k in K ], 'k-')
        ax.plot( x, [ (1.0*i_lst[k][1])/i_lst[k][0] for k in K ], 'k:')
        ax.plot( x, [ (1.0*i_lst[k][2])/i_lst[k][0] for k in K ], 'k-.')
        ax.plot( x, [ (1.0*i_lst[k][3])/i_lst[k][0] for k in K ], 'k--')
        #ax.set_xticks( [] )
        ax.legend( lgnd, 'lower right' )
        ax.set_title( 'Number of iterations(p)/Number of iterations(0)' )
        ax.set_xticklabels( K, rotation = 45, horizontalalignment = 'right' )
        MM.show()
        
