"""
Simple demo of the pygltr module.
A number of Ampl models or nl files are read on
the command line, and two versions of gltr are called,
with the same trust-region radius. A comparison of the
running times for the two versions is illustrated on a
simple plot.
"""
import amplpy
import pygltr
from pysparse import spmatrix
import numpy
import norms
from nlpy_timing import cputime
import sys

def ReadNlFile( fname ):
    """
    Read Ampl model or nl graph from file fname.
    Returns an AmplModel object, or None if an error occured.
    """
    nlp = amplpy.AmplModel( fname )
    return nlp

def gltr_explicit_fromfile( fname, g, radius ):
    nlp = ReadNlFile( fname )
    if nlp is None: return None
    return gltr_explicit( nlp, g, radius )

def gltr_explicit( nlp, g, radius ):
    H = nlp.hess( nlp.x0, nlp.pi0 )
    G = pygltr.PyGltrContext( g, radius = radius )
    t = cputime()
    G.explicit_solve( H.to_csr() )
    t = cputime() - t
    return (G.m, G.mult, G.snorm, G.niter, G.nc, G.ierr, t)

def gltr_implicit_fromfile( fname, g, radius ):
    nlp = ReadNlFile( fname )
    if nlp is None: return None
    return gltr_implicit( nlp, g, radius )

def gltr_implicit( nlp, g, radius ):
    G = pygltr.PyGltrContext( g, radius = radius )
    t = cputime()
    G.implicit_solve( lambda v: nlp.hprod( nlp.pi0, v ) )
    t = cputime() - t
    return (G.m, G.mult, G.snorm, G.niter, G.nc, G.ierr, t)
    

# Test the module
if __name__ == '__main__':

    # Problems are specified on the command line
    ProbList = sys.argv[1:]
    nprobs = len( ProbList )
    if nprobs < 1:
        sys.stderr.write( 'Please specifiy one or more problem names\n' )
        sys.exit( -3 )

    radius = 10.0
    t_list_I  = []
    t_list_II = []

    print 'Problem name               Size       #nnz       M I        M II' 
    print '-' * 69

    for p in range( len( ProbList ) ):
        problem = ProbList[p]
        nlp = ReadNlFile( problem )
        if nlp is not None:
            g = numpy.ones( nlp.n, 'd' )
            (f, m, sn, nit, nc, ierr, t1) = gltr_explicit( nlp, g, radius )
            t_list_I.append( t1 )
            (f, m, sn, nit, nc, ierr, t2) = gltr_implicit( nlp, g, radius )
            t_list_II.append( t2 )
            
            print '%-25s  %-9d  %-9d  %-9g  %-9g' % (problem,nlp.n,nlp.nnzh,t1,t2)
            if problem[-3:] == '.nl':
                ProbList[p] = problem[:-3]
            nlp.close()
            #del nlp
        else:
            # Remove problem from list
            ProbList[p] = []
            nprobs -= 1

    print '-' * 69
    print
    print ' Method I  : pass Hessian matrix explicitly'
    print ' Method II : pass function for Hessian-vector products'
    print

    # Plot the timings
    try:
        import matplotlib
        if matplotlib.__version__ < '0.65':
            import matplotlib.matlab as MM
        else:
            import matplotlib.pylab as MM
    except:
        print ' If you had Matplotlib installed, you would be'
        print ' looking at timing plots right now.'
        sys.exit( 0 )

    darkblue = '#2c11cf'
    lightblue = '#8f84e0'
    x = range( nprobs )
    ax = MM.axes()
    ax.plot( x, t_list_I,  color = darkblue, linewidth = 3 )
    ax.plot( x, t_list_II, color = lightblue, linewidth = 3 )
    ax.set_xticks( x )
    ax.set_xticklabels( ProbList, rotation = 45, horizontalalignment = 'right' )
    ax.set_ylabel( 'Time (s)' )
    ax.set_title( 'Comparing Explicit and Implicit versions of PyGLTR' )
    ax.legend( ['Explicit', 'Implicit'], 'upper left' )
    MM.show()
