"""
Illustrates usage of PyMa27 for factorization of Symmetric Quasi-Definite (sqd)
matrices. For a description and properties of sqd matrices, see

 R. J. Vanderbei, Symmetric Quasi-Definite Matrices,
 SIAM Journal on Optimization, 5(1), 100-113, 1995.

D. Orban, Montreal, December 2007.
$Id:$
"""

import pyma27
from pysparse import spmatrix
from numpy import empty, ones, linalg, Inf
from nlpy_timing import cputime
import pylab
import sparsetools
import sys

# Create symmetric quasi-definite matrix K
A = spmatrix.ll_mat_from_mtx( 'Data/bcsstk18.mtx' )
C = spmatrix.ll_mat_from_mtx( 'Data/bcsstk11.mtx' )

nA = A.shape[0]
nC = C.shape[0]
K = spmatrix.ll_mat_sym(nA + nC, A.nnz + C.nnz + min(nA,nC))
K[:nA,:nA] = A
K[nA:,nA:] = C
K[nA:,nA:].scale(-1.0)
for i in range(min(nA,nC)): K[nA+i,i] = 1.0

# Create right-hand side rhs=K*e
e = ones(nA+nC)
rhs = empty(nA+nC)
K.matvec(e,rhs)

# Factorize and solve Kx = rhs, knowing K is sqd
t = cputime()
P = pyma27.PyMa27Context(K, sqd = True)
t = cputime() - t
sys.stderr.write('Factorization time with sqd=True : %5.2fs   ' % t )
P.solve(rhs, get_resid=False)
sys.stderr.write('Error: %7.1e\n' % linalg.norm(P.x - e, ord=Inf))

# Do it all over again, pretending we don't know K is sqd
t = cputime()
P = pyma27.PyMa27Context(K, sqd = False)
t = cputime() - t
sys.stderr.write('Factorization time with sqd=False: %5.2fs   ' % t )
P.solve(rhs, get_resid=False)
sys.stderr.write('Error: %7.1e\n' % linalg.norm(P.x - e, ord=Inf))

# Plot sparsity pattern
fig = pylab.figure()
cur_ax = fig.gca()
cur_ax.plot( [0,nA+nC-1], [nA,nA], 'b--', linewidth=1 )
cur_ax.plot( [nA,nA], [0,nA+nC-1], 'b--', linewidth=1 )
sparsetools.spymatll(K, ax=cur_ax)
pylab.show()
