from pysparse import spmatrix
from sparsetools import usolve, utsolve, lsolve, ltsolve, uppertriangularize, lowertriangularize
import numpy
import norms
from nlpy_timing import cputime
import sys
import os

# Illustrate the solution of upper and lower triangular systems
# The system solver is written in C.
# For comparisons, look at demo_triangular_sys_python, in which
# the system solver is written in Python.
#
# D. Orban                                   Montreal, Oct 2005

if __name__ == '__main__':
    matrices = sys.argv[1:]

    header = '%-15s  %-6s  %-8s  %-12s  %-6s  %-6s  %-12s  %-6s  %-6s\n' % ('Matrix', 'n', 'nnz', '|Mx-b|', 'Me', 'usol', "|M'x-b|", "M'e", 'utsol' )
    l = len( header )
    sys.stdout.write( header )
    sys.stdout.write( '-' * l + '\n' )

    for matrix in matrices:
        M = spmatrix.ll_mat_from_mtx( matrix )

        probname = os.path.basename( matrix )
        if probname[-4:] == '.mtx': probname = probname[:-4]
        
        n = M.shape[0]
        e = numpy.ones( n, 'd' )
        b = numpy.zeros( n, 'd' )
        uppertriangularize( M )

        sys.stdout.write( '%-15s  %-6d  %-8d' % (probname, n, M.nnz) )

        # Construct b = U e
        t = cputime()
        M.matvec( e, b )
        t_matvec = cputime() - t

        # Solve U x = b. Solution is e.
        t = cputime()
        x = usolve( M, b )
        t_usolve = cputime() - t
        sys.stdout.write( '  %-12g' % norms.norm_infty(x-e) )
        sys.stdout.write( '  %-6g  %-6g' % (t_matvec, t_usolve) )

        # Construct b = U' e
        t = cputime()
        M.matvec_transp( e, b )
        t_matvec_transp = cputime() - t

        # Solve U' x = b. Solution is e.
        t = cputime()
        x = utsolve( M, b )
        t_utsolve = cputime() - t
        sys.stdout.write( '  %-12g' % norms.norm_infty(x-e) )
        sys.stdout.write( '  %-6g  %-6g\n' % (t_matvec_transp, t_utsolve) )

        # Same operation, with lower triangle

        M = spmatrix.ll_mat_from_mtx( matrix )
        lowertriangularize( M )
        sys.stdout.write( '%-33s' % '' )

        # Construct b = L e
        t = cputime()
        M.matvec( e, b )
        t_matvec = cputime() - t

        # Solve L x = b. Solution is e.
        t = cputime()
        x = lsolve( M, b )
        t_lsolve = cputime() - t
        sys.stdout.write( '  %-12g' % norms.norm_infty(x-e) )
        sys.stdout.write( '  %-6g  %-6g' % (t_matvec, t_lsolve) )

        # Construct b = L' e
        t = cputime()
        M.matvec_transp( e, b )
        t_matvec_transp = cputime() - t

        # Solve L' x = b. Solution is e.
        t = cputime()
        x = ltsolve( M, b )
        t_ltsolve = cputime() - t
        sys.stdout.write( '  %-12g' % norms.norm_infty(x-e) )
        sys.stdout.write( '  %-6g  %-6g\n' % (t_matvec_transp, t_ltsolve) )

        
    sys.stdout.write( '-' * l + '\n' )
