# This module implements a basic steepest descent
# method with Armijo backtracking linesearch.

from numpy import dot

class SteepestDescent:
    """
    Perform a steepest descent method with linesearch to solve the
    unconstrained minimization problem
    
        minimize   f(x).

    The model should be described as an instance of the NLPModel class
    and the linesearch should be an instance of the LineSearch class.
    
    For example, if nlp is an NLPModel instance either built by hand
    or read from an AMPL model and if ALS is an ArmijoLinesearch,
    we simply issue the call

        SD = SteepestDescent( nlp, ALS )
        SD.solve()

    Recognized keyword arguments that may be passed upon instantiation
    are

        maxiter    the maximal number of iterations allowed.
    """
    def __init__( self, nlp, linesearch, **kwargs ):

        # Save reference to model and linesearch
        self.nlp = nlp
        self.linesearch = linesearch
        self.best_estimate = None
        self.iter = 0
        self.optimal = False

        if 'maxiter' in kwargs.keys():
            self.maxiter = kwargs['maxiter']
        else:
            self.maxiter = 15000

    def solve( self ):

        x     = self.nlp.x0
        grad  = self.nlp.grad(x)
        d     = -grad
        iter  = 0
        
        while not self.nlp.AtOptimality( x, None, gradient = grad ) and iter < self.maxiter:

            iter += 1
            slope = dot( grad, d )
            f     = self.nlp.obj(x)

            # Perform linesearch
            t = self.linesearch.search( self.nlp.obj, x, d, slope, f = f )
    
            # Move to new iterate
            x   += t * d
            grad = self.nlp.grad( x )
            d    = -grad

        self.best_estimate = x
        self.iter = iter
        if self.iter < self.maxiter: self.optimal = True
        # And voila!


# Demo usage of class SteepestDescent

if __name__ == '__main__':

    import numpy
    from rosenbrock import RosenbrockModel
    from linesearch import ArmijoLineSearch

    # Adjust starting point and
    # stopping tolerance if desired
    RosenbrockModel.x0 = numpy.array( [-0.5, 1.0], 'd' )
    RosenbrockModel.stop_d = 10.0**(-5)

    # Define linesearch method
    ALS = ArmijoLineSearch( beta = 1.0e-4, tfactor = 0.1 )

    # Instantiate steepest descent algorithm
    SD = SteepestDescent( RosenbrockModel, ALS, maxiter = 10000 )

    # Minimize the model
    SD.solve()

    # Output final info
    print ' Optimal solution found: ', SD.optimal
    print ' Best estimate found   : ', SD.best_estimate
    print ' Value at best estimate: ', RosenbrockModel.obj( SD.best_estimate )
    print ' Number of iterations  : ', SD.iter
